<?php

namespace Services;

use Doctrine\ORM\EntityManager as EntityManager;
use Util\FactoryInstance;
use Util\Utilities;
use Util\InstanceProvider;
use Util\RepositoryInstance;
use Exception\Login as Exception;
use Teknisa\Libs\Service\Menu;
use Teknisa\Libs\Util\Utilities as LibUtilities;
use Teknisa\Model\Geral\Entities\OperadorAcesso;
use Teknisa\Model\Geral\Entities\Operador;

class Login
{
    const IS_ACTIVE_POSITIVE = 'Y';
    const DATE_TYPE_DAY = 'D';
    const ID_RESET_PASSWORD = 'R';
    const USES_ACTIVE_DIRECTORY_POSITIVE = 'Y';
    const USES_PASSWORD_CONTROL_POSITIVE = 'Y';
    const USES_IP_RESTRICTION_POSITIVE = 'Y';
    const RECURRENCE_DAYS = 7;

    const BLOCKED_PASSWORD_STATUS = 'B';
    const INVALID_PASSWORD_STATUS = 'I';

    const RESET_PASSWORD_IDENTIFICATION   = 'RESET_PASSWORD';
    const EXPIRED_PASSWORD_IDENTIFICATION = 'EXPIRED_PASSWORD';
    const PASSWORD_WILL_EXPIRE_IDENTIFICATION = 'PASSWORD_WILL_EXPIRE';
    const PASSWORD_EXPIRE_TODAY_IDENTIFICATION = 'PASSWORD_EXPIRE_TODAY';
    const IS_ADM_USER = 'S';

    protected $loginUrl;
    protected $loginClientSecret;

    protected $libClientId;
    protected $libClientSecret;

    protected $licenseExceptions;

    public function __construct()
    {
        $config = LibUtilities::getLoginConfig();
        $this->loginUrl          = $config['loginUrl'];
        $this->loginClientSecret = $config['loginClientSecret'];

        $integration = LibUtilities::getLoginIntegration();
        $this->libClientId     = $integration['clientId'];
        $this->libClientSecret = $integration['clientSecret'];

        $this->licenseExceptions = array (
            22, /** Product is not available */
            23, /** Menu items not found */
            24  /** Product date expired */
        );
    }

    /**
     * @param $user
     * @param $password
     * @param $productId
     * @param $requesterUrl
     * @param $attempts
     * @param $fullOperator
     * @param $hash
     * @param $sessionChange
     * @param $keepConnected
     * @param $rcUrl
     * @return array
     * @throws \Exception
     */
    public function login($user, $password, $productId, $requesterUrl, $attempts, $fullOperator, $hash, $sessionChange, $keepConnected, $rcUrl) {
        $userData     = InstanceProvider::getLibServiceApi()->getUserData($user, $password);
        $hasUserData  = !empty($userData['USER']);
        $teknisaOrgId = null;

        if ($hasUserData) {
            /** @var NRORGOPER or NRORG $organizationId */
            $organizationId = $this->getOrganizationId($userData);
            /** @var NRORGOPER or CUSTOM_NRORG or NRORG $teknisaOrgId */
            $teknisaOrgId  = $this->getTeknisaOrganizationId($userData);
            /** @var NRORG $defaultOrgId */
            $defaultOrgId  = $userData['USER']['ORGANIZATION_ID'];
            $userId        = $userData['USER']['ID'];
            $userName      = $userData['USER']['NAME'];
            $nrParcNegocio = $userData['USER']['NRPARCNEGOCIO'];
            $cdFilial      = $userData['USER']['CDFILIALSELECIONADA'];
            if ($nrParcNegocio && $userId) {
                $idUtlPerfilFilial = $fullOperator ? RepositoryInstance::getOperador()->getIdUtlPerfilFilialParamGeral($defaultOrgId) : null;
                if (empty($idUtlPerfilFilial) || $idUtlPerfilFilial['IDUTLPERFILFILIAL'] != 'S') {
                    $userData['ACCESS_PROFILE'] = RepositoryInstance::getOperador()->getOperatorAccessProfile($nrParcNegocio, $defaultOrgId);
                } else {
                    $selectedProfile = RepositoryInstance::getOperador()->getCdPerfilSelecionado($defaultOrgId, $userId);
                    if (empty($selectedProfile) || empty($selectedProfile[0]['ID'])) {
                        $userData['ACCESS_PROFILE'] = RepositoryInstance::getOperador()->getBranchAccessProfile($userId, $defaultOrgId);
                    } else {
                        $userData['ACCESS_PROFILE'] = $selectedProfile;
                    }
                    $admUser = $this->verifyAdministratorUser($userData['ACCESS_PROFILE'], $userData['OTHER_SESSION_VALUES']['IDUTCTRLACESWEB']);
                    $userData['OTHER_SESSION_VALUES']['ACL_ADMINISTRATOR'] = $admUser;
                }
                $customAccessProfile = $this->getAccessProfileByExternalFunction($defaultOrgId, $userId);
                $userData['ACCESS_PROFILE'] = is_null($customAccessProfile) ? $userData['ACCESS_PROFILE'] : $customAccessProfile;
            } else {
                $userData['ACCESS_PROFILE'] = [];
            }
        }

        $expiredPassword = $this->validations($userData, $password, $productId, $requesterUrl, $attempts);

        $sessionId = InstanceProvider::getSession()->getId();
        $token = InstanceProvider::getOAuth()->grantAccessToken($this->libClientId, $this->libClientSecret, array('sessionId' => $sessionId));

        // Check if exists concurrent access to user
        if($hash && $keepConnected && !$this->isPasswordOperSupp($password)) {
            $this->checkConcurrentAccess($productId, $hash, $userId, $defaultOrgId, $sessionChange, $token, $keepConnected);
        }

        $filter = $this->getLoginParams($teknisaOrgId, $productId, $user, $password, $hasUserData);
        try {
            $loginResponse = Utilities::callWebService(null, '/productLicense', $filter, 'userData', $this->loginClientSecret, $this->loginUrl);
        } catch(\Exception $e) {
            if(in_array($e->getCode(), $this->licenseExceptions)) {
                throw $e;
            }
            $loginResponse = null;
        }

        if (!$hasUserData && !$loginResponse) {
            Exception::invalidPassword();
        }

        $isSupportOperator = empty($loginResponse['IS_SUPPORT_OPERATOR']) ? false : true;
        $userData = $isSupportOperator ? $loginResponse['USER_DATA'] : $userData;
        if ($this->isPasswordOperSupp($password)) {
            $this->validateWebPassword($userData['USER'], $password, false, $attempts, null, true);
        }

        if($isSupportOperator) {
            $alertExpirationDatePer = false;
            $userId = $loginResponse['USER_DATA']['USER']['ID'];
            $oAccCustomFile = Utilities::getOAccCustomParameter();
            if (file_exists($oAccCustomFile)) {
                $paramgeral = RepositoryInstance::getParamgeral()->findParamgeralWithoutNrorg();
                if(!empty($paramgeral)) {
                    $defaultOrgId = $paramgeral['NRORG'];
                    $loginResponse['USER_DATA']['USER']['NRORG'] = $paramgeral['NRORG'];
                    $loginResponse['USER_DATA']['USER']['NRORGTRAB'] = $paramgeral['NRORG'];
                    $loginResponse['USER_DATA']['USER']['ORGANIZATION_ID'] = $paramgeral['NRORG'];
                } else {
                    $defaultOrgId = $loginResponse['USER_DATA']['USER']['NRORGSUP'];
                }
            } else {
                $defaultOrgId = $loginResponse['USER_DATA']['USER']['NRORGSUP'];
            }

            $loginResponse['USER_DATA']['USER']['CDFILIALSELECIONADA'] = null;
            $this->createSupporteOperator($loginResponse['USER_DATA']['USER'], $defaultOrgId);
            $userData = $loginResponse['USER_DATA'];
        } else {
            $alertExpirationDatePer = isset($loginResponse['EXPIRATION_DATE_PER']) ? $this->validatePer($loginResponse['EXPIRATION_DATE_PER']) : false;
        }

        if(empty($loginResponse)) {
            $productPermission = $this->getProductPermissions($productId, $organizationId, $defaultOrgId);
            $loginResponse = array(
                "TOKEN"              => null,
                "REDIRECT_URL"       => null,
                "REDIRECT_TO_RC"     => null,
                "ERROR_CALL_LICENSE" => true,
                "VISIBLE_ELEMENTS"   => $productPermission['elements'],
                "EXPIRED_ELEMENTS"   => isset($productPermission['expiredElements']) ? $productPermission['expiredElements'] : array(),
                "REPOSITORIES_INFO"  => $productPermission['migrationRepositories']
            );
        } else {
            $loginResponse['ERROR_CALL_LICENSE'] = false;
            if (!empty($loginResponse['ORG_IS_ACTIVE']) && $loginResponse['ORG_IS_ACTIVE'] === 'N') {
                Exception::inactiveOrganization();
            }
        }

        if(!$isSupportOperator) {
            // Set File Server API Key in Login Response, according to parameters.
            $loginResponse = $this->setFSAPIKeyInLoginResponse($loginResponse, $teknisaOrgId);
        }

        return $this->factorySessionData($userData, $loginResponse, $expiredPassword, $alertExpirationDatePer, $isSupportOperator, $rcUrl, $token, $teknisaOrgId);
    }

    private function createSupporteOperator($user, $defaultOrgId) {
        $entityManager = InstanceProvider::getLibConnectionCustom()->updateVpdConnection($defaultOrgId);
        $entityManager = $entityManager ? $entityManager : InstanceProvider::getEntityManager();

        $entityOperRepository = $entityManager->getRepository(Utilities::GENERAL_ENTITY_NAMESPACE . 'Operador');
        $operadorRepository = RepositoryInstance::getOperador();
        $operadorFactory = FactoryInstance::getOperador();

        $suppOperator = $entityOperRepository->findOneBy(array('cdoperador' => $user['ID'], 'nrorg' => $defaultOrgId));
        $suppOperator = $operadorFactory->saveSupportOperator($suppOperator, $defaultOrgId, $user['PASSWORD_MD5'], $user['ID'],
            $user['NAME'], $user['EMAIL']);
        $entityManager->persist($suppOperator);
        $entityManager->flush();

        $operadorRepository->insertBranchesSupportOperator($defaultOrgId, $user['ID']);

        try {
            $operatorCode = $user['ID'];
            $operatorName = $user['SUP_NAME'];
            $supOrgId     = $user['NRORGSUP'];
            $observation  = "Login como Operador de suporte NMOPERADOR: $operatorName e NRORG: $supOrgId";
            $searchItem   = "CDOPERADOR=$operatorCode NRORG=$defaultOrgId";
            $host         = isset($_SERVER['HTTP_HOST']) ? $_SERVER['HTTP_HOST'] : $_SERVER['SERVER_NAME'];
            $connParams   = $entityManager->getConnection()->getParams();
            $dbUserName   = isset($connParams['sessionMode']) ? $connParams['dbname'] : $connParams['user'];
            $operadorRepository->insertLogOpera('1', $observation, $searchItem, gethostbyname($host), $dbUserName, $defaultOrgId, $operatorCode, $operatorName);
        } catch(\Exception $e) { }
    }

    private function checkConcurrentAccess($productId, $hash, $userId, $organizationId, $sessionChange, $token, $keepConnected) {
        $entityManager = InstanceProvider::getLibConnectionCustom()->updateVpdConnection($organizationId);
        $entityManager = $entityManager ? $entityManager : InstanceProvider::getEntityManager();

        $dateNow = new \DateTime();
        $tokenLifeTime = LibUtilities::getTokenLifeTimeParameter();
        if (\Teknisa\Libs\Util\Utilities::getConcurrentAccessSaveInMongoParameter()) {
            $this->accessMongo($productId, $hash, $userId, $organizationId, $sessionChange, $token, $keepConnected,
                $entityManager, $dateNow, $tokenLifeTime);
        } else {
            $this->accessOracle($productId, $hash, $userId, $organizationId, $sessionChange, $token, $keepConnected,
                $entityManager, $dateNow, $tokenLifeTime);
        }
    }

    private function isValidPreviousToken(OperadorAcesso $concurrentAccess) {
        $isValidToken  = true;
        $keepConnected = $concurrentAccess->getIdmanterconectado() == 'S';
        try {
            InstanceProvider::getMongoCache()->fetch($concurrentAccess->getToken(), $concurrentAccess->getNrtempovidatoken(), $keepConnected);
        } catch(\Exception $e) {
            $isValidToken = false;
        }
        return $isValidToken;
    }

    private function isValidPreviousTokenMongo(array $concurrentAccess) {
        $isValidToken  = true;
        try {
            $keepConnected = empty($concurrentAccess['idmanterconectado']) ? false : $concurrentAccess['idmanterconectado'] == 'S';
            $token = $concurrentAccess['token'];
            $nrtempovidatoken = $concurrentAccess['nrtempovidatoken'];
            InstanceProvider::getMongoCache()->fetch($token, $nrtempovidatoken, $keepConnected);
        } catch(\Exception $e) {
            $isValidToken = false;
        }
        return $isValidToken;
    }

    /**
     * @param $userData
     * @param $loginResponse
     * @param $expiredPassword
     * @param $alertExpirationDatePer
     * @return array
     */
    private function factorySessionData($userData, $loginResponse, $expiredPassword, $alertExpirationDatePer, $isSupportOperator, $rcUrl, $token, $teknisaOrgId) {
        $user = $userData['USER'];
        $userInfo = array(
            "USER_ID"                   => $user['ID'],
            "USER_EMAIL"                => $user['EMAIL'],
            "TOKEN"                     => $token,
            "REDIRECT_URL"              => $loginResponse['REDIRECT_URL'],
            "REDIRECT_TO_RC"            => $loginResponse['REDIRECT_TO_RC'],
            "VISIBLE_ELEMENTS"          => $loginResponse['VISIBLE_ELEMENTS'],
            "EXPIRED_ELEMENTS"          => isset($loginResponse['EXPIRED_ELEMENTS']) ? $loginResponse['EXPIRED_ELEMENTS'] : [],
            "REPOSITORIES_INFO"         => $loginResponse['REPOSITORIES_INFO'],
            "USER_DATA"                 => $userData,
            "FILE_SERVER_API_KEY"       => isset($loginResponse["FILE_SERVER_API_KEY"]) ? $loginResponse["FILE_SERVER_API_KEY"] : null,
            "IS_SUPPORT_OPERATOR"       => $isSupportOperator
        );

        $encryptedSession = LibUtilities::zeedhiEncrypt(json_encode($userInfo));
        $licenseInfo = false;

        if(!$loginResponse['ERROR_CALL_LICENSE']) {
            $licenseInfo = array(
                'RELATED_ORG_IDS'     => isset($loginResponse['RELATED_ORG_IDS']) ? $loginResponse['RELATED_ORG_IDS'] : array(),
                'RELATED_ORGS'        => isset($loginResponse['RELATED_ORGS']) ? $loginResponse['RELATED_ORGS'] : array(),
                'ORGS_AND_BRANCHES'   => isset($loginResponse['ORGANIZATION_AND_BRANCHES_IDS']) ? $loginResponse['ORGANIZATION_AND_BRANCHES_IDS'] : array(),
                'PRODUCTS_PERMISSION' => $loginResponse['PRODUCTS_PERMISSION'],
                'VARIABLE_MEASURE'    => $loginResponse['VARIABLE_MEASURE'],
                'PRIMARY_ID'          => $loginResponse['PRIMARY_ID'],
                'GENERAL_COLLECTION'  => $loginResponse['GENERAL_COLLECTION'],
                "LICENSE_TOKEN"       => $loginResponse['TOKEN'],
                "LICENSE_URL"         => $this->loginUrl,
                "LICENSE_SECRET"      => $this->loginClientSecret,
                "USER_ID"             => $user['ID'],
                "USER_NAME"           => $user['NAME'],
                "TEKNISA_ORG_ID"      => $teknisaOrgId,
            );
            $licenseInfo = LibUtilities::zeedhiEncrypt(json_encode($licenseInfo));
        }

        $sessionData = array(
            "USER_ID"    => $user['ID'],
            "USER_NAME"  => $user['NAME'],
            "USER_EMAIL" => $user['EMAIL'],

            "REDIRECT_URL"     => $loginResponse['REDIRECT_URL'],
            "REDIRECT_TO_RC"   => $loginResponse['REDIRECT_TO_RC'],
            "TOKEN"            => $token,
            "USER"             => $encryptedSession,
            "EXPIRED_PASSWORD" => $expiredPassword,
            "LOGGED"           => true,
            "LICENSE_INFO"     => $licenseInfo,

            "IS_SUPPORT_OPERATOR"       => $isSupportOperator,
            "ALERT_EXPIRATION_DATE_PER" => $alertExpirationDatePer
        );

        if((!empty($loginResponse['REDIRECT_TO_RC']) && !empty($rcUrl)) || (!empty($loginResponse['REDIRECT_URL']))) {
            $this->setRedirectionProductData($encryptedSession, $token, $licenseInfo);
        }

        return $sessionData;
    }

    /**
     * @param $productId
     * @param $organizationId
     * @param $defaultOrgId
     * @return array
     * @throws \Exception
     */
    public function getProductPermissions($productId, $organizationId, $defaultOrgId)
    {
        $oAccTekMgr = RepositoryInstance::getOAccTekMgr()->getOAccTekMgrByNrOrg($organizationId, $defaultOrgId);
        if(empty($oAccTekMgr)) {
            Exception::productIsNotAuthorized();
        }
        $oAccTekDecrypted = LibUtilities::zeedhiDecrypt($oAccTekMgr['OACCTEK'], LibUtilities::LOGIN_CRYPT_STRING);

        $orgAccess = json_decode($oAccTekDecrypted, true);
        if (empty($orgAccess)) {
            Exception::orgAccessDecryptFail();
        }

        foreach ($orgAccess['products'] as $orgProduct) {
            if (isset($orgProduct['productId']) && $orgProduct['productId'] == $productId) {
                $productPermission = $orgProduct;
                break;
            }
        }

        if(empty($productPermission)) {
            Exception::productIsNotAvailable();
        }
        if(empty($productPermission['elements'])) {
            Exception::menuItemsNotFound();
        }

        return $productPermission;
    }

    /**
     * @param $password
     * @param $isSupportOperator
     * @return array
     */
    public function validateUser($password, $isSupportOperator) {
        $user = InstanceProvider::getLibEnvironment()->getCurrentUserId();
        if($isSupportOperator) {
            $this->validateUserInLicense($user, $password);
        } else {
            $userData = array();
            $userDataProvider = InstanceProvider::getLibUserDataProvider();

            $operator = InstanceProvider::getLibUserDataRepository()->getOperatorByCdoperador($user);
            $userDataProvider->getPasswordParameters($operator['CDOPERADOR'], $operator['NRORG'], $userData);
            $passwordParams = $userData['PASSWORD_PARAMETERS'];

            if(!$userDataProvider->validateAD($passwordParams, $operator['NMACESSO'], $password, $operator['NRORG'])) {
                $this->passwordValidation($operator, $password, $passwordParams, false);
            }

        }
        InstanceProvider::getLibEnvironment()->setSessionExpired(false);
        return array(true);
    }

    public function validateUserInLicense($cdoperador, $password)
    {
        $params =  array (
            "requestType" => "FilterData",
            "filter" => array(
                array (
                    'name' => 'CDOPERADOR', 'operator' => '=', 'value' => $cdoperador
                ), array (
                    'name' => 'PASSWORD', 'operator' => '=', 'value' => $password
                )
            )
        );

        $validatedUser = Utilities::callWebService(null, '/validateUserSupportOperator',
                                   $params, 'validateUserSupportOperator',
                                   $this->loginClientSecret, $this->loginUrl);
        if(!$validatedUser['EXPIRATION_DATE']) {
            Exception::expiredTokenSupportOperator();
        }

        if(!$validatedUser['PASSWORD']) {
            Exception::invalidPasswordCode32();
        }
    }

    /**
     * @param $organizationId
     * @param $productId
     * @param $user
     * @param $password
     * @param $hasUserData
     * @return array
     */
    protected function getLoginParams($organizationId, $productId, $user, $password, $hasUserData) {
        $menus = InstanceProvider::getLibMenuService()->buildMenu(array(Menu::ALL_ELEMENTS_VISIBLE), null);
        $elementIds = array();
        $this->getElementIds($menus, $elementIds);
        return array (
            "requestType" => "FilterData",
            "filter" => array( array(
                'name' => 'ORGANIZATION_ID', 'operator' => '=', 'value' => $organizationId
            ), array (
                'name' => 'PRODUCT_ID',      'operator' => '=', 'value' => $productId
            ), array (
                'name' => 'ELEMENT_IDS',     'operator' => '=', 'value' => $elementIds
            ), array (
                'name' => 'SERVER_ID',       'operator' => '=', 'value' => Utilities::getServerIdParameter()
            ), array (
                'name' => 'USER',            'operator' => '=', 'value' => $user
            ), array (
                'name' => 'PASSWORD',        'operator' => '=', 'value' => $password
            ), array (
                'name' => 'HAS_USER_DATA',   'operator' => '=', 'value' => $hasUserData
            ), array (
                'name' => 'RETURN_ORG_AND_BRANCHES', 'operator' => '=', 'value' => true
            ), array (
                'name' => 'USE_SUPPORT_OPERATOR', 'operator' => '=', 'value' => Utilities::getUseSupportOperatorParameter()
            ) )
        );
    }

    /**
     * @param $menus
     * @param $elementIds
     */
    protected function getElementIds($menus, &$elementIds) {
        $menus = isset($menus['MENUS']) ? $menus['MENUS'] : $menus;
        foreach($menus as $menu) {
            if(isset($menu['id'])) {
                array_push($elementIds, $menu['id']);
            }
            if(isset($menu['items'])) {
                $this->getElementIds($menu['items'], $elementIds);
            }
        }
    }

    /**
     * @param $userData
     * @param $password
     * @param $productId
     * @param $requesterUrl
     * @param $attempts
     * @return bool|string
     * @throws \Exception
     */
    protected function validations($userData, $password, $productId, $requesterUrl, $attempts) {
        if (!$this->isPasswordOperSupp($password)) {
            if(empty($userData) || empty($userData['USER'])) {
                Exception::invalidPassword();
            }
            $otherSessionValues = $userData['OTHER_SESSION_VALUES'];
            $passwordStatus = $this->userValidation($userData['USER'], $productId);
            $this->accessProfileValidation($userData['ACCESS_PROFILE'], $otherSessionValues);
            $this->validateIp($userData['IP_RESTRICTION'], LibUtilities::getCurrentIp());
            $this->validateDomain($requesterUrl);

            $expiredPassword = $this->passwordValidation($userData['USER'], $password, $userData['PASSWORD_PARAMETERS'],
                true, $attempts, $otherSessionValues['ACL_ADMINISTRATOR'], $passwordStatus);

            return $expiredPassword;
        }
    }

    protected function isPasswordOperSupp($password) {
        return strpos($password, 'TekOperSupp_') !== false;
    }

    protected function validatePer($expirationDatePer)
    {
        $alertExpirationDatePer = false;
        if ($expirationDatePer) {
            $dtAtual = new \DateTime();
            $datetime = new \DateTime();
            $validadePer = $datetime::createFromFormat('d/m/Y H:i:s', "$expirationDatePer 23:59:59");
            if($validadePer > $dtAtual) {
                $dateInterval = $validadePer->diff($dtAtual);
                if ($dateInterval->days <= 30) {
                    $alertExpirationDatePer = $dateInterval->days;
                }
            } else {
                Exception::expirationDatePer();
            }
        }
        return $alertExpirationDatePer;
    }

    /**
     * @param $user
     * @param $productId
     */
    protected function userValidation($user, $productId) {
        if(!isset($user['IS_ACTIVE']) || $user['IS_ACTIVE'] !== self::IS_ACTIVE_POSITIVE) {
            Exception::userInactive();
        }
        if(empty($user['ORGANIZATION_ID'])) {
            Exception::userNotOrganization();
        }
        if (empty($productId)) {
            Exception::productIdNotFound();
        }

        $operatorParams = RepositoryInstance::getOperador()->getOperatorParams($user['ID'], $user['ORGANIZATION_ID']);
        if($operatorParams['IDSTATUSENHA'] == self::INVALID_PASSWORD_STATUS) {
            Exception::accessBlocked();
        }

        return $operatorParams['IDSTATUSENHA'];
    }

    /**
     * @param $accessProfile
     * @param $otherSessionValues
     */
    protected function accessProfileValidation($accessProfile, $otherSessionValues) {
        if(empty($accessProfile) &&
            ((isset($otherSessionValues['IDUTCTRLACESWEB']) && $otherSessionValues['IDUTCTRLACESWEB'] == 'S') ||
                (isset($otherSessionValues['IDUTCTRLACES']) && $otherSessionValues['IDUTCTRLACES']    == 'S'))
        ) {
            Exception::userNotAccessProfile();
        }
    }

    /**
     * @param $user
     * @param $password
     * @param $passwordParams
     * @param bool $isValidateExpired
     * @param $attempts
     * @param $isAdmin
     * @return bool|string
     */
    protected function passwordValidation($user, $password, $passwordParams, $isValidateExpired = true, $attempts = null, $isAdmin = null, $passwordStatus = null) {
        $isAD = $passwordParams && $passwordParams['USES_ACTIVE_DIRECTORY'] == self::USES_ACTIVE_DIRECTORY_POSITIVE;

        /** If exists attempts, this method called in login.
         *  Else, called by expired session */
        if(empty($attempts)) {
            $paramSenha = null;
        } else {
            $entityManager = InstanceProvider::getLibConnectionCustom()->updateVpdConnection($user['ORGANIZATION_ID']);
            $paramSenha    = RepositoryInstance::getParamSenha()->getParamSenhaByNrOrg($user['ORGANIZATION_ID'], $entityManager);
        }

        if(!$user['PASSWORD_MD5']) {
            $this->validateDelphiPassword($user, $password, $isAD, $attempts, $paramSenha, $isAdmin);
        } else {
            $this->validateWebPassword($user, $password, $isAD, $attempts, $paramSenha, $isAdmin);
        }

        if($passwordStatus && $passwordStatus == self::BLOCKED_PASSWORD_STATUS) {
            return self::EXPIRED_PASSWORD_IDENTIFICATION;
        }

        if($isValidateExpired && $passwordParams) {
            return $this->validateExpiredPassword($passwordParams, $paramSenha);
        }

        return false;
    }

    /**
     * @param $user
     * @param $loginPassword
     * @param $isAD
     * @param $attempts
     * @param $paramSenha
     * @param $isAdmin
     * @throws \Exception
     */
    protected function validateDelphiPassword($user, $loginPassword, $isAD, $attempts, $paramSenha, $isAdmin) {
        $loginPasswordEncrypted = !$isAD ? LibUtilities::getDelphiPasswordEncrypted($loginPassword) : $loginPassword;
        if($user['PASSWORD_CRYPTO_PLUS'] != $loginPasswordEncrypted) {
            $this->throwInvalidPasswordError($user['ID'], $user['ORGANIZATION_ID'], $attempts, $paramSenha, $isAdmin);
        }
    }

    /**
     * @param $user
     * @param $loginPassword
     * @param $isAD
     * @param $attempts
     * @param $paramSenha
     * @param $isAdmin
     * @throws \Exception
     */
    protected function validateWebPassword($user, $loginPassword, $isAD, $attempts, $paramSenha, $isAdmin) {
        $loginPasswordEncrypted = !$isAD ? LibUtilities::encrypt($loginPassword) : $loginPassword;
        if($user['PASSWORD_MD5'] != $loginPasswordEncrypted) {
            $this->throwInvalidPasswordError($user['ID'], $user['ORGANIZATION_ID'], $attempts, $paramSenha, $isAdmin);
        }
    }

    /**
     * @param $userId
     * @param $organizationId
     * @param $attempts
     * @param $paramSenha
     * @param $isAdmin
     * @throws \Exception
     */
    public function throwInvalidPasswordError($userId, $organizationId, $attempts, $paramSenha, $isAdmin) {
        if(empty($paramSenha) || $isAdmin) {
            Exception::invalidPassword();
        } else {
            $nrErrors = $paramSenha[0]['NRERROSENHA'];
            if(!empty($nrErrors) && $attempts == $nrErrors - 1) {
                Exception::invalidPasswordBlockUser();
            } else if(!empty($nrErrors) && $attempts >= $nrErrors) {
                RepositoryInstance::getOperador()->saveOperatorPasswordStatus($userId, $organizationId, self::INVALID_PASSWORD_STATUS);
                Exception::accessBlockedByAttempts();
            } else {
                Exception::invalidPasswordCode32();
            }
        }
    }

    /**
     * @param $passwordParams
     * @param $paramSenha
     * @return bool|string
     */
    protected function validateExpiredPassword($passwordParams, $paramSenha) {
        if($passwordParams['IDENTIFICATION'] == self::ID_RESET_PASSWORD) {
            return self::RESET_PASSWORD_IDENTIFICATION;
        }
        if(empty($passwordParams['UPDATE_DATE'])) {
            return self::EXPIRED_PASSWORD_IDENTIFICATION;
        }

        if(!empty($passwordParams['PASSWORD_EXPIRATION_DAYS']) && $passwordParams['USES_PASSWORD_CONTROL'] == self::USES_PASSWORD_CONTROL_POSITIVE) {
            $expireDate = \DateTime::createFromFormat('d/m/Y', $passwordParams['UPDATE_DATE']);
            $interval = 'P' . ($passwordParams['PASSWORD_EXPIRATION_DAYS']) . self::DATE_TYPE_DAY;
            $expireDate->add(new \DateInterval($interval));
            $actualDate = new \DateTime();
            $intervalExp  = Utilities::getDateIntervalInSeconds($expireDate, $actualDate);
            $intervalDays = Utilities::getDateIntervalInSeconds($actualDate, $expireDate);
            if($intervalDays > 0) {
                $intervalDays = floor($intervalDays / 86400);
            } else if($intervalDays < 0) {
                $intervalDays = null;
            }
            if ($intervalExp > 0) {
                return self::EXPIRED_PASSWORD_IDENTIFICATION;
            } else if($intervalDays == 0) {
                return self::PASSWORD_EXPIRE_TODAY_IDENTIFICATION;
            } else if($intervalDays && $intervalDays <= $paramSenha[0]['NRDIASAVISO']) {
                return self::PASSWORD_WILL_EXPIRE_IDENTIFICATION . '_' . $intervalDays;
            }
        }
        return false;
    }

    /**
     * @param $ipRestriction
     * @param $ip
     * @throws \Exception
     */
    protected function validateIp($ipRestriction, $ip) {
        if(!empty($ipRestriction) && $ipRestriction['USES_IP_RESTRICTION'] == self::USES_IP_RESTRICTION_POSITIVE) {
            $isValidIp = false;
            if(Utilities::isIpv4($ip)) {
                foreach($ipRestriction['IPV4'] as $ipV4) {
                    if(Utilities::validateIpInRange($ip, $ipV4)) {
                        $isValidIp = true;
                        break;
                    }
                }
            } else {
                foreach($ipRestriction['IPV6'] as $ipV6) {
                    if($ip == $ipV6) {
                        $isValidIp = true;
                        break;
                    }
                }
            }
            if(!$isValidIp) {
                Exception::ipIsOutOfRange($ip);
            }
        }
    }

    /**
     * @param $url
     * @return bool
     */
    protected function validateDomain($url) {
        if(Utilities::getUseDomainFilterParameter()) {
            $validDomains = explode(';', Utilities::getDomainFilterParameter());
            foreach($validDomains as $domain) {
                if(!empty($domain) && strpos($url, $domain) !== false) {
                    return true;
                }
            };
            Exception::invalidDomain();
        }
        return true;
    }

    public function getProductLanguages($productId) {
        $filter = array(
            "requestType" => "FilterData",
            "filter" => array( array(
                'name' => 'PRODUCT_ID', 'operator' => '=', 'value' => $productId
            ) )
        );
        return Utilities::callWebService(null, '/productLanguages', $filter, 'languages', $this->loginClientSecret, $this->loginUrl);
    }

    /**
     * @return array
     * @throws \Exception\CustomException
     */
    protected function readOAccCustom() {
        $oAccCustom = array();
        try {
            $fileDir = Utilities::getOAccCustomParameter();
            if (file_exists($fileDir)) {
                $fileContents  = file_get_contents($fileDir);
                $fileDecrypted = LibUtilities::zeedhiDecrypt($fileContents, LibUtilities::LOGIN_CRYPT_STRING);
                $fileDecoded   = @json_decode($fileDecrypted, true);
                if(isset($fileDecoded['organizationId']) &&
                    isset($fileDecoded['dbHost']) &&
                    is_numeric($fileDecoded['organizationId'])) {
                    $oAccCustom = $fileDecoded;
                }
            }
        } catch(\Exception $e) {}

        if(!empty($oAccCustom)) {
            if(Utilities::getDbHostInfo() != $oAccCustom['dbHost']) {
                Exception::incorrectDbHostInfo();
            }
        };

        return $oAccCustom;
    }

    /**
     * @param $organizationId
     */
    public function genOAccCustom($organizationId) {
        $dbHost = Utilities::getDbHostInfo();
        if(!empty($organizationId) && !empty($dbHost)) {
            $oAccCustom = array(
                'organizationId' => $organizationId,
                'dbHost' => $dbHost
            );
            $oAccCrypted = LibUtilities::zeedhiEncrypt(json_encode($oAccCustom), LibUtilities::LOGIN_CRYPT_STRING);
            file_put_contents(Utilities::getOAccCustomParameter(), $oAccCrypted);
            echo "File created!\n";
        } else {
            if(empty($organizationId)) echo "Organization ID not found.\n";
            if(empty($dbHost)) echo "DB Host Data not found.\n";
        }
    }

    public function validateRedirectionUrl($redirectUrl) {
        $url = str_replace('/mobile', '', $redirectUrl) . '/backend/index.php';
        $url = 'http://localhost/pm-next2/backend/service/index.php';
        $result = Utilities::callWebService(null, '/lib_testConnection', array(), 'connection', $this->libClientSecret, $url, false, true);
        if(empty($result)) {
            throw new \Exception('Erro na requisição da URL de redirecionamento');
        }
        return $result;
    }

    private function setRedirectionProductData($user, $token, $licenseInfo) {
        $dataset = array('$set' => array (
            'TOKEN' => $token,
            'USER'  => $user,
            'LICENSE_INFO' => $licenseInfo,
        ));
        $filter = array(
            'TOKEN' => $token
        );
        InstanceProvider::getMongoDB()->update('LoginRedirection', $filter, $dataset, true);
    }


    private function getAccessProfileByExternalFunction($nrOrg, $operatorCode) {
        $function = Utilities::getAccessProfileFunctionParameter();
        if(empty($function)) {
            return null;
        } else {
            return InstanceProvider::getEntityManager()->getRepository(Utilities::GENERAL_ENTITY_NAMESPACE . 'Operador')
                ->$function($nrOrg, $operatorCode);
        }
    }

    public function testLicenseConnection() {
        $error = false;
        try {
            Utilities::callWebService(null, '/testConnection', array(), 'testConnection',
                $this->loginClientSecret, $this->loginUrl, false, true);
        } catch(\Exception $e) {
            $error = $e->getMessage();
        }
        return array (
            'ERROR' => $error,
            'URL'   => explode('/backend', $this->loginUrl)[0]
        );
    }

    private function getTeknisaOrganizationId($userData) {
        $oAccCustom     = $this->readOAccCustom();
        $nrOrgOper      = $userData['USER']['NRORGOPER'];
        $organizationId = $userData['USER']['ORGANIZATION_ID'];

        if(!empty($nrOrgOper)) {
            return $nrOrgOper;
        }
        if(!empty($oAccCustom)) {
            return $oAccCustom['organizationId'];
        }
        return $organizationId;
    }

    private function getOrganizationId($userData) {
        $nrOrgOper      = $userData['USER']['NRORGOPER'];
        $organizationId = $userData['USER']['ORGANIZATION_ID'];

        if(!empty($nrOrgOper)) {
            return $nrOrgOper;
        }
        return $organizationId;
    }

    private function setFSAPIKeyInLoginResponse($loginResponse, $teknisaOrgId) {
        try {
            if(empty($loginResponse['FILE_SERVER_API_KEY'])) { // Don't have a key.
                if(Utilities::getEnableSaveFSAPIkeyInDBParameter())
                    $loginResponse['FILE_SERVER_API_KEY'] = InstanceProvider::getFileServerService()->getAPIKey($teknisaOrgId); // Get it from the database.
                else
                    $loginResponse['FILE_SERVER_API_KEY'] = null; // There's no key avaiable.
            } else {
                if(Utilities::getEnableSaveFSAPIkeyInDBParameter()) {
                    InstanceProvider::getFileServerService()->saveAPIKeyInDB($loginResponse['FILE_SERVER_API_KEY'], $teknisaOrgId); // Save key in DB.
                }
            }
        } catch(\Exception $e) {}

        return $loginResponse;
    }


    private function saveNewAccess($productId, $hash, $userId, $organizationId, $token, $keepConnected,
                                   \Factory\OperadorAcesso $operadorAcessoFactory, OperadorAcesso $concurrentAccess = null,
                                   \DateTime $dateNow, $tokenLifeTime, EntityManager $entityManager): void
    {
        $connection = $entityManager->getConnection();
        $operadorAcessoFactory->save($concurrentAccess, $productId, $hash, $userId, $organizationId,
            $dateNow, $token, $tokenLifeTime, $keepConnected, $connection);
        $entityManager->flush();
    }

    private function accessOracle($productId, $hash, $userId, $organizationId, $sessionChange, $token,
                                  $keepConnected, $entityManager, \DateTime $dateNow, $tokenLifeTime): void
    {
        $operadorAcessoFactory = FactoryInstance::getOperadorAcesso();

        /** @var OperadorAcesso $concurrentAccess */
        $concurrentAccess = $entityManager->getRepository(Utilities::GENERAL_ENTITY_NAMESPACE . 'OperadorAcesso')
            ->findOneBy(array("nrprodutoId" => $productId, "cdoperador" => $userId, "nrorg" => $organizationId));
        if (!$concurrentAccess || !$concurrentAccess->getDsChaveUltAcesso()) {
            $this->saveNewAccess($productId, $hash, $userId, $organizationId, $token, $keepConnected,
                $operadorAcessoFactory, $concurrentAccess, $dateNow, $tokenLifeTime, $entityManager);
        } else {
            if ($concurrentAccess->getDsChaveUltAcesso() != $hash) {
                if ($sessionChange || ($concurrentAccess->getToken() && !$this->isValidPreviousToken($concurrentAccess))) {
                    $this->saveNewAccess($productId, $hash, $userId, $organizationId, $token, $keepConnected,
                        $operadorAcessoFactory, $concurrentAccess, $dateNow, $tokenLifeTime, $entityManager);
                } else {
                    Exception::concurrentAccess();
                }
            }
        }
    }

    private function saveNewAccessMongo($productId, $hash, $userId, $organizationId, $token, $keepConnected,
                                    array $concurrentAccess = null,\DateTime $dateNow, $tokenLifeTime)
    {
        $operadorAcessoFactory = FactoryInstance::getOperadorAcessoMongo();
        $concurrentAccess = $operadorAcessoFactory->save($concurrentAccess, $productId, $hash, $userId, $organizationId,
            $dateNow, $token, $tokenLifeTime, $keepConnected);
        InstanceProvider::getMongoCache()->saveAccess($concurrentAccess);
    }

    private function accessMongo($productId, $hash, $userId, $organizationId, $sessionChange, $token,
                                 $keepConnected, $entityManager, \DateTime $dateNow, $tokenLifeTime)
    {
        /** @var array $concurrentAccess */
        $concurrentAccess = InstanceProvider::getMongoCache()->findAccess(
            array("nrProdutoId" => intval($productId), "cdOperador" => $userId, "nrOrg" => $organizationId)
        );
        if (empty($concurrentAccess) || empty($concurrentAccess['dsChaveUltAcesso'])) {
            $this->saveNewAccessMongo($productId, $hash, $userId, $organizationId, $token, $keepConnected,
                 $concurrentAccess, $dateNow, $tokenLifeTime);
        } else {
            if ($concurrentAccess['dsChaveUltAcesso'] != $hash) {
                if ($sessionChange || (!empty($concurrentAccess['Token']) && !$this->isValidPreviousTokenMongo($concurrentAccess))) {
                    $this->saveNewAccessMongo($productId, $hash, $userId, $organizationId, $token, $keepConnected,
                        $concurrentAccess, $dateNow, $tokenLifeTime);
                } else {
                    Exception::concurrentAccess();
                }
            }
        }
    }

    protected function verifyAdministratorUser($accessProfile, $idutctrlacesweb) {
        $admUser = false;
        if($idutctrlacesweb == 'N') {
            $admUser = true;
        } else {
            foreach ($accessProfile as $profile) {
                if ($profile['IDPERFILACESSOADM'] == self::IS_ADM_USER) {
                    $admUser = true;
                    break;
                }
            }
        }
        return $admUser;
    }

}
